# Copyright (c) 2021 Andrei Sukhanov. All rights reserved.
#
# Licensed under the MIT License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://github.com/ASukhanov/upstrim/blob/main/LICENSE
#
__version__ = "1.0.3"



# Copyright (c) 2021 Andrei Sukhanov. All rights reserved.
#
# Licensed under the MIT License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://github.com/ASukhanov/apstrim/blob/main/LICENSE
#

"""Base class for accessing multiple Data Objects, served by a liteServer.
#``````````````````Usage:`````````````````````````````````````````````````````
ipython3
import pupLITE as LA 
from pprint import pprint

# If the name resolution using liteCNS is not configured, then
# replace Scaler1 with an IP address of the the host name of the liteScaler. 
LAserver = 'Scaler1:server'
LAdev1   = 'Scaler1:dev1'
LAdev2   = 'Scaler1:dev2'
    #``````````````Info```````````````````````````````````````````````````````
    # list of all devices on a server
#DNW#pprint(list(LA.PVs([[['Scaler1','*'],'*']]).info()))
    # info on all parameters of the 'Scaler1','server'
pprint(LA.PVs((LAserver,'*')).info())
    # info on single parameter
pprint(LA.PVs((LAserver,'timeShift')).info())
    # info on multiple parameters
#DNW#pprint(LA.PVs((LAserver,('timeShift','perf')).info())
    #``````````````Get```````````````````````````````````````````````````````
    # get all parameters from device LAserver
pprint(LA.PVs((LAserver,'*')).get())
    # get single parameter from device Scaler1:server:
pprint(LA.PVs((LAserver,    'perf')).get()) # or:
pprint(LA.PVs((('Scaler1','server'),'perf')).get())
    # get multiple parameters from device Scaler1:server: 
pprint(LA.PVs((LAserver,('perf','timeShift'))).get())
    # get multiple parameters from multiple devices 
pprint(LA.PVs((LAdev1,('time','frequency')),(LAdev2,('time','coordinate'))).get())
    # simplified get: returns (value,timestamp) of a parameter 'perf' 
pprint(LA.PVs((LAserver,'perf')).value)
    #``````````````Read```````````````````````````````````````````````````````
    # get all readable parameters from device Scaler1:server, which have been modified since the last read
print(LA.PVs((LAserver,'*')).read())#TODO
    #``````````````Set````````````````````````````````````````````````````````
    # simplified set, for single parameter:
LA.PVs((LAdev1,'frequency')).value = [1.1]
    # explicit set, could be used for multiple parameters:
LA.PVs((LAdev1,'frequency')).set([1.1])
pprint(LA.PVs([[['Scaler1','dev1'],'frequency']]).value)
    # multiple set
LA.PVs([[['Scaler1','dev1'],['frequency','coordinate']]]).set([8.,[3.,4.]])
LA.PVs([[['Scaler1','dev1'],['frequency','coordinate']]]).get()
    #``````````````Subscribe``````````````````````````````````````````````````
ldo = LA.PVs([[['Scaler1','dev1'],'cycle']])
ldo.subscribe()# it will print image data periodically
ldo.unsubscribe()# cancel the subscruption

#``````````````````Programmatic way, using Access`````````````````````````````
# Advantage: The previuosly created PVs are reused
LAserver = 'Scaler1:server'
LA.Access.get((LAserver,'*'))
LA.Access.subscribe(LA.testCallback,(LAdev1,'cycle'))
LA.Access.subscribe(LA.testCallback,(LAdev2,'time'))

    # test for timeout, should timeout in 10s:
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#''''''''''''''''''Using Access interface`````````````````````````````````````
LA.Access.subscribe(LA.testCallback,(LAdev1,('cycle','time')))
LA.Access.unsubscribe()

#``````````````````TODO``````````````````````````````````````````````````````` 
#ldo = LA.PVs([[['localhost','dev1'],'*']])
#ldo.subscribe()
#LA.PVs(['Scaler1.dev1','frequency']).set(property=('oplimits',[-1,11])
#``````````````````Observations```````````````````````````````````````````````
    # Measured transaction time is 1.8ms for:
LA.PVs([[['Scaler1','dev1'],['frequency','command']]]).get()
    # Measured transaction time is 6.4ms per 61 KBytes for:
LA.PVs([[['Scaler1','dev1'],'*']]).read() 
#``````````````````Tips```````````````````````````````````````````````````````
To enable debugging: LA.PVs.Dbg = True
To enable transaction timing: LA.Channel.Perf = True  
"""
#__version__ = 'v64 2021-05-04'# raising exceptions instead of returning error code.
__version__ = 'v65 2021-06-10'# subscription sockets are blocking now


print('liteAccess '+__version__)

import sys, time, socket
from os import getpid
import getpass
from timeit import default_timer as timer
import threading
recvLock = threading.Lock()

#from pprint import pformat, pprint
import ubjson

#````````````````````````````Globals``````````````````````````````````````````
UDP = True
Port = 9700
PrefixLength = 4
socketSize = 1024*64 # max size of UDP transfer
Dev,Par = 0,1
NSDelimiter = ':'# delimiter in the name field

Username = getpass.getuser()
Program = sys.argv[0]
PID = getpid()
#print(f'liteAccess user:{Username}, PID:{PID}, program:{Program}')
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#````````````````````````````Helper functions`````````````````````````````````
MaxPrint = 500
def croppedText(obj, limit=200):
    txt = str(obj)
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt
def printTime(): return time.strftime("%m%d:%H%M%S")
def printi(msg): 
    print(croppedText(f'INFO.LA@{printTime()}: '+msg))

def printw(msg):
    msg = msg = croppedText(f'WARN.LA@{printTime()}: '+msg)
    print(msg)
    #Device.setServerStatusText(msg)

def printe(msg): 
    msg = croppedText(f'ERROR.LA@{printTime()}: '+msg)
    print(msg)
    #Device.setServerStatusText(msg)

def printd(msg):
    if PVs.Dbg or Access.Dbg : print('Dbg.LA: '+msg)

def croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt

def testCallback(args):
    printi(croppedText(f'>testCallback({args})'))

def printCallback(args):
    print(f'subcribed item received:\n{args}')

def ip_address():
    """Platform-independent way to get local host IP address"""
    return [(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close())\
        for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]

CNSMap = {}# local map of cnsName to host:port
def _hostPort(cnsNameDev:tuple):
    """Return host;port of the cnsName,Dev try it first from already 
    registered records, or from the name service"""
    global CNSMap
    if len(cnsNameDev) == 1:
        printe(f'Device name wrong: {cnsNameDev}, should be of the form: host:dev')
        sys.exit(1)
    cnsName,dev = cnsNameDev
    if isinstance(cnsName,list):
        cnsName = tuple(cnsName)
    # printd(f'>_hostPort: {cnsNameDev}')
    try:  
        hp,dev = CNSMap[cnsName]# check if cnsName is in local map
    except  KeyError:
        from . import liteCNS
        printi(f'cnsName {cnsName} not in local map: {CNSMap}')
        try:
            hp = liteCNS.hostPort(cnsName)
        #except NameError:
        except Exception as e:
            msg = (f'The host name {cnsName} is not in liteCNS: {e}\n'
                f"Trying to use it as is: '{cnsName}'")
            #raise   NameError(msg)
            printw(msg)
            hp = cnsName
        # register externally resolved cnsName in local map
        hp = hp.split(';')
        hp = tuple(hp) if len(hp)==2 else (hp[0],Port)            
        #printi('cnsName %s is locally registered as '%cnsName+str((hp,dev)))
        CNSMap[cnsName] = hp,dev
        printi(f'Assuming host,port: {hp}')
    except ValueError:
        printe(f'Device name {cnsName} wrong, it should be in form host:device ')
        sys.exit(1)
    h,p = hp
    try:
        h = socket.gethostbyname(h)
    except:
        printe(f'Could not resolve host name {h}') 
    return h,p

retransmitInProgress = None
def _recvUdp(sock, socketSize):
    """Receive the chopped UDP data"""
    port = sock.getsockname()[1]
    #print(f'>_recvUdp {port} locked: {recvLock.locked()}')
    #with recvLock:
    global retransmitInProgress
    chunks = {sock:{}}
    tryMore = 5# Max number of allowed lost packets
    ts = timer()
    ignoreEOD = 3

    def ask_retransmit(offsetSize):
        global retransmitInProgress
        retransmitInProgress = tuple(offsetSize)
        cmd = {'cmd':('retransmit',offsetSize)}
        printi(f'Asking to retransmit port {port}: {cmd}')
        sock.sendto(ubjson.dumpb(cmd),addr)
    
    while tryMore:
        try:
            buf, addr = sock.recvfrom(socketSize)
        #else:#except Exception as e:
        except socket.timeout as e:
            msg = f'Timeout in recvfrom port {port}'
            printi(msg)
            raise
        if buf is None:
            raise RuntimeError(msg)
        size = len(buf) - PrefixLength
        offset = int.from_bytes(buf[:PrefixLength],'big')# python3
        
        #DNPprinti(f'chunk received at port {port}: {offset,size}')

        if size > 0:
            chunks[sock][offset,size] = buf[PrefixLength:]
        if offset > 0 and not retransmitInProgress:
            # expect more chunks to come
            continue

        # check if chunk is EOD, i.e. offset,size = 0,0
        if size == 0:
            ignoreEOD -= 1
            if ignoreEOD >= 0:
                #print(f'premature EOD{ignoreEOD} received from, ignore it')
                continue
            else:
                msg = f'Looks like first chunk is missing at port {port}'
                printw(msg)
                #This is hard to recover. Give up
                return [],0
        else:
            #print('First chunk received')
            pass

        if retransmitInProgress is not None:
            if (offset,size) in chunks[sock]:
                printi(f'retransmission received {offset,size}')
            else:
                printw(f'server failed to retransmit chunk {retransmitInProgress}')
                tryMore = 1
            retransmitInProgress = None

        # last chunk have been received, offset==0, size!=0
        # check for lost  chunks
        sortedKeys = sorted(chunks[sock])
        prev = [0,0]
        allAssembled = True
        for offset,size in sortedKeys:
            #print('check offset,size:'+str((offset,size)))
            last = prev[0] + prev[1]
            if last != offset:
                l = offset - last
                if l > 65536:
                    msg = f'Lost too many bytes at port {port}: {last,l}, data discarded'
                    printw(msg)
                    #raise RuntimeError(msg)
                    return [],0
                    #return 'WARNING: '+msg, addr
                ask_retransmit((last, l))
                allAssembled = False
                break
            prev = [offset,size]

        if allAssembled:
            break
        #print(f'tryMore: {tryMore}')
        tryMore -= 1
    ts1 = timer()
        
    if not allAssembled:
        msg = 'Partial assembly of %i frames'%len(chunks[sock])
        #raise BufferError(msg)
        printw(msg)
        return [],0
        #return ('WARNING: '+msg).encode(), addr

    data = bytearray()
    sortedKeys = sorted(chunks[sock])
    for offset,size in sortedKeys:
        # printd('assembled offset,size '+str((offset,size)))
        data += chunks[sock][(offset,size)]
    tf = timer()
    # if len(data) > 500000:
        # printd('received %i bytes in %.3fs, assembled in %.6fs'\
        # %(len(data),ts1-ts,tf-ts1))
    #printi('assembled %i bytes'%len(data))
    return data, addr

def send_dictio(dictio, sock, hostPort:tuple):
    """low level send"""
    global LastDictio
    LastDictio = dictio.copy()
    # printd('executing: '+str(dictio))
    dictio['username'] = Username
    dictio['program'] = Program
    dictio['pid'] = PID
    # printd(f'send_dictio to {hostPort}: {dictio}')
    encoded = ubjson.dumpb(dictio)
    if UDP:
        sock.sendto(encoded, hostPort)
    else:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect(hostPort)
        except Exception as e:
            printe('in sock.connect:'+str(e))
            sys.exit()
        sock.sendall(encoded)

def send_cmd(cmd, devParDict:dict, sock, hostPort:tuple, values=None):
    import copy
    #DNPprint(f'send_cmd({cmd}.{devParDict}')
    if cmd == 'set':
        if len(devParDict) != 1:
            raise ValueError('Set is supported for single device only')
        for key in devParDict:
            devParDict[key] += 'v',values
    devParList = list(devParDict.items())
    dictio = {'cmd':(cmd,devParList)}
    # printd('sending cmd: '+str(dictio))
    send_dictio(dictio, sock, hostPort)

def llTransaction(dictio, sock, hostPort:tuple):
    # low level transaction
    send_dictio(dictio, sock, hostPort)
    return  receve_dictio()

def receive_dictio(sock, hostPort:tuple):
    '''Receive and decode message from associated socket'''
    # printd('\n>receive_dictio')
    if UDP:
        data, addr = _recvUdp(sock, socketSize)
        # acknowledge the receiving
        sock.sendto(b'ACK', hostPort)
    #printd('received %i of '%len(data)+str(type(data))+' from '+str(addr)':')
    # decode received data
    # allow exception here, it will be caught in execute_cmd
    if len(data) == 0:
        printw(f'empty reply for: {LastDictio}')
        return {}
    try:
        decoded = ubjson.loadb(data)
    except Exception as e:
        printw(f'exception in ubjson.load Data[{len(data)}]: {e}')
        #print(str(data)[:150])
        #raise ValueError('in receive_dictio: '+msg)
        return {}
    #for key,val in decoded.items():
    #    print(f'received from {key}: {val.keys()}')

    if not isinstance(decoded,dict):
        #print('decoded is not dict')
        return decoded
    for cnsDev in decoded:
        # items could be numpy arrays, the following should decode everything:
        parDict = decoded[cnsDev]
        for parName,item in list(parDict.items()):
            # printd(f'parName {parName}: {parDict[parName].keys()}')
            # check if it is numpy array
            shapeDtype = parDict[parName].get('numpy')
            if shapeDtype is not None:
                #print(f'par {parName} is numpy {shapeDtype}')
                shape,dtype = shapeDtype
                v = parDict[parName]['value']
                # it is numpy array
                from numpy import frombuffer
                parDict[parName]['value'] =\
                    frombuffer(v,dtype).reshape(shape)
                del parDict[parName]['numpy']
            else:
                #print(f'not numpy {parName}')
                pass
    # printd(f'<receive_dictio')
    return decoded

class Subscriber():
    def __init__(self, hostPort:tuple, devParDict:dict, callback):
        self.name = f'{hostPort,devParDict}'
        self.hostPort = hostPort
        self.devParDict = devParDict
        self.callback = callback

class SubscriptionSocket():
    event = threading.Event()
    '''handles all subscriptions to single hostPort'''
    def __init__(self, hostPort):
        #printi(f'>subsSocket {hostPort}')
        self.name = f'{hostPort}'
        self.hostPort = tuple(hostPort)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        #self.socket.bind(self.hostPort)
        self.callback = None# holds the callback for checking
        
        dispatchMode = 'Thread'
        if dispatchMode == 'Thread':
            self.selector = None
            #self.socket.settimeout(10)
            #self.thread = thread_with_exception(target=self.receivingThread, args=[])
            self.thread = threading.Thread(target=self.receivingThread)
            self.thread.daemon = True
            self.event.clear()
            self.thread.start()

    def receivingThread(self):
        printi(f'>receiving thread started for {self.hostPort}') 
        while not self.event.is_set():
            try:
                dictio = receive_dictio(self.socket, self.hostPort)
            #except socket.timeout as e:
            #except RuntimeError as e:
            except Exception as e:
                msg = f'in subscription thread socket {self.name}: '+str(e)
                printw(msg)
                raise
                #dictio = {'WARNING':msg}
                #dictio = None
            self.dispatch(dictio)
        printi(f'<receiving thread stopped for {self.hostPort}') 
        
    def subscribe(self, subscriber):
        if self.socket is None:
            print(f'UDP socket created for {self.hostPort}')
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # just checking:
        if subscriber.hostPort != self.hostPort:
            printe(f'Subscribe logic error in {self.name}')# this should never happen
        #printd(f'subscribing {subscriber.name}')
        if self.callback is None:
            self.callback = subscriber.callback
        else:
            if self.callback != subscriber.callback:
                printe(f'Only one callback is supported per hostPort, subscription for {subscriber.name} is discarded')
                return
        self.event.clear()
        send_cmd('subscribe', subscriber.devParDict, self.socket\
        , self.hostPort)

    def unsubscribe_all(self):
        printi(f'unsubscribing {self.hostPort}, {self.socket}')
        if self.socket is None:
            return
        send_cmd('unsubscribe', {'*':'*'}, self.socket, self.hostPort)
        printi(f'killing thread of {self.name}')
        self.event.set()
        print(f'shutting down {self.hostPort}')
        try:    self.socket.shutdown(socket.SHUT_RDWR)
        except Exception as e: 
            printw(f'Exception in shutting down: {e}')
        print(f'closing down {self.hostPort}')
        try:    self.socket.close()
        except Exception as e: 
            printw(f'Exception in closing: {e}')
        self.socket = None

    def dispatch(self, dictio):
        if dictio:
            #printi(croppedText(f'>dispatch {dictio}'))
            self.callback(dictio)
        else:
            printw(f'empty data from {self.hostPort}')
            return
    
subscriptionSockets = {}

def add_subscriber(hostPort:tuple, devParDict:dict, callback=testCallback):
    subscriber = Subscriber(hostPort, devParDict, callback)
    #self.subscribers[subscriber.name] = subscriber

    # register new socket if not registered yet
    try:
        ssocket = subscriptionSockets[hostPort]
    except:
        ssocket = SubscriptionSocket(hostPort)
        subscriptionSockets[hostPort] =  ssocket
        #printd(f'new socket in publishingHouse: {hostPort}')
    subscriptionSockets[hostPort] = ssocket
    ssocket.subscribe(subscriber)

def unsubscribe_all():
    global subscriptionSockets
    for hostPort,ssocket in subscriptionSockets.items():
        ssocket.unsubscribe_all()
    subscriptionSockets = {}
    printi('all unsibscribed')

#TODO: cache the channel sockets for reuse
pvSockets = {}
class Channel():
    Perf = False
    """Provides access to host;port"""#[(dev1,[pars1]),(dev2,[pars2]),...]
    def __init__(self, hostPort:tuple, devParDict={}, timeout=10):
        # printd(f'>Channel {hostPort,devParDict}')
        self.devParDict = devParDict
        host = hostPort[0]
        if host.lower() in ('','localhost'):
            host = ip_address()
        try:    port = int(hostPort[1])
        except: port = 9700
        self.hostPort = host,port
        self.name = f'{self.hostPort}'
        self.recvMax = 1024*1024*4
        if UDP:
            self.sock = pvSockets.get(self.hostPort)
            if self.sock is None:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                self.sock.settimeout(timeout)
                pvSockets[self.hostPort] = self.sock
                
        #print('%s client of %s, timeout %s'
        #%(('TCP','UDP')[UDP],str((self.sHost,self.sPort)),str(timeout)))

    def _llTransaction(self, dictio):
        # low level transaction
        send_dictio(dictio, self.sock, self.hostPort)
        return receive_dictio(self.sock, self.hostPort)

    def _sendCmd(self, cmd, values=None):
        r = send_cmd(cmd, self.devParDict, self.sock, self.hostPort, values)

    def _transaction(self, cmd, value=None):
        # normal transaction: send command, receive response
        if Channel.Perf: ts = timer()
        self._sendCmd(cmd,value)
        r = receive_dictio(self.sock, self.hostPort)
        if Channel.Perf: print('transaction time: %.5f'%(timer()-ts))
        # printd(f'reply from channel {self.name}: {r}')
        return r
    
class PVs(object): #inheritance from object is needed in python2 for properties to work
    """Class, representing multiple data access objects."""
    Dbg = False
    subscriptionsCancelled = True

    def __init__(self, *ldoPars, timeout=5):
        self.timeout = timeout
        # printd(f'``````````````````Instantiating PVs ldoPars:{ldoPars}')        
        # unpack arguments to hosRequest map
        self.channelMap = {}
        if isinstance(ldoPars[0], str):
            printe('Device,parameter should be a list or tuple')
            return
            sys.exit(1)
        for ldoPar in ldoPars:
            ldo = ldoPar[0]
            pars = ldoPar[1] if len(ldoPar) > 1 else ''
            try:    ldo = ldo.split(NSDelimiter)
            except: pass
            ldo = tuple(ldo)
            # printd(f'ldo,Pars:{ldo,pars}')            
            if isinstance(pars,str): pars = [pars]
            # ldo is in form: (hostName,devName)
            ldoHost = _hostPort(ldo)
            #cnsNameDev = ','.join(ldoPar[0])
            cnsNameDev = NSDelimiter.join(ldo)
            #print('ldoHost,cnsNameDev',ldoHost,cnsNameDev)
            if ldoHost not in self.channelMap:
                self.channelMap[ldoHost] = {cnsNameDev:[pars]}
                # printd(f'created self.channelMap[{ldoHost,self.channelMap[ldoHost]}')
            else:
                try:
                    # printd(f'appending old cnsNameDev {ldoHost,cnsNameDev} with {pars[0]}')
                    self.channelMap[ldoHost][cnsNameDev][0].append(pars[0])
                except:
                    # printd(f'creating new cnsNameDev {ldoHost,cnsNameDev} with {pars[0]}')
                    self.channelMap[ldoHost][cnsNameDev] = [pars]
                #print(('updated self.channelMap[%s]='%ldoHost\
                #+ str(self.channelMap[ldoHost]))
        channelList = list(self.channelMap.items())
        # printd(f',,,,,,,,,,,,,,,,,,,channelList constructed: {channelList}')
        self.channels = [Channel(*i) for i in channelList]
        return

    def devices(self):
        """Return list of devices on associated host;port"""
        return self.channels[0]._llTransaction({'cmd':['info']})

    def info(self):
        for channel in self.channels:
            return channel._transaction('info')

    def get(self):
        for channel in self.channels:
            return channel._transaction('get')

    def _firstValueAndTime(self):
        if True:#try: skip 'server,device'
            firstDict = self.channels[0]._transaction('get')
            if not isinstance(firstDict,dict):
                return firstDict
            firstValsTDict = list(firstDict.values())[0]
        else:#except Exception as e:
            printw('in _firstValueAndTime: '+str(e))
            return (None,)
        # skip parameter key
        firstValsTDict = list(firstValsTDict.values())[0]
        ValsT = list(firstValsTDict.values())[:2]
        try:     return (ValsT[0], ValsT[1])
        except:  return (ValsT[0],)

    #``````````````Property 'value````````````````````````````````````````````
    # It is for frequently needed get/set access to a single parameter
    @property
    def value(self):
        """Request from server first item of the PVs and return its 
        value and timestamp,"""
        return self._firstValueAndTime()

    @value.setter
    def value(self,value):
        """Send command to set the value to the first item of the PVs"""
        return self.set(value)
    #,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,

    def read(self):
        """Return only readable parameters"""
        for channel in self.channels:
            return channel._transaction('read')

    def set(self,value):
        #for channel in self.channels:
        #TODO: the set is not supported yet for multiple objects
        for channel in self.channels[:1]:
            r = channel._transaction('set',value)
            if isinstance(r,str):
                raise RuntimeError(r)
        return r

    #``````````````subscription ``````````````````````````````````````````````
    def subscribe(self, callback=testCallback):
        if len(self.channels) > 1:
            raise NameError('subscription is supported only for single host;port')
        channel = self.channels[0]
        add_subscriber(channel.hostPort, channel.devParDict, callback)

    def unsubscribe(self):
        unsubscribe_all()

#``````````````````Universal Access```````````````````````````````````````````
#PVC_PV, PVC_CB, PVC_Props = 0, 1, 2
PVC_PV, PVC_Thread = 0, 1
    
class Access():
    """Universal interface to liteServer parameters, similar to cad_io.
    The pvName should be a tuple: (deviceName,parameterName)
    The full form of the device name: hostName;Port:deviceName,
    if Port is default then it can be omitted: hostName:deviceName.
    Returned values are remapped to the form {(hostDev,par):{parprop:{props}...}...}
    that could be time consuming for complex requests.
    """
    _Subscriptions = []
    __version__ = __version__
    Dbg = False

    def info(*devParNames):
        return PVs(*devParNames).info()

    def get(*devParNames, **kwargs):# kwargs are for compatibility with ADO, they ignored here
        return PVs(*devParNames).get()

    def set(devPar_Value):
        #Only one object supported so far
        dev,par,value = devPar_Value
        return PVs((dev,par)).set(value)

    def subscribe(callback, *devParNames):
        if not callable(callback):
            printe(('subscribe arguments are wrong,'
            'expected: subscribe(callback,(dev,par))'))
            return
        pvs = PVs(*devParNames)
        pvs.subscribe(callback)

    def unsubscribe():
        '''Unsubscribe all paramters'''
        unsubscribe_all()
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,



# Copyright (c) 2021 Andrei Sukhanov. All rights reserved.
#
# Licensed under the MIT License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://github.com/ASukhanov/apstrim/blob/main/LICENSE
#
#__version__ = '1.0.3 2021-06-01'# EPICS and LITE support is OK, Compression supported
#__version__ = '1.0.4a 2021-06-11'# flush the file after each section
#__version__ = '1.0.4 2021-06-11'# if file exist then rename the existing file
#__version__ = '1.0.5 2021-06-14'# handling of different returned maps
__version__ = '1.0.6 2021-06-19'# filename moved from instantiation to new method: start(), timestamp is int(nanoseconds)

import sys, time, string, copy
import os, pathlib, datetime
import threading
import signal

import numpy as np
import msgpack
import msgpack_numpy
msgpack_numpy.patch()

#````````````````````````````Globals``````````````````````````````````````````
SecDateTime, SecParagraph = 0,1
Nano = 1000000000
#````````````````````````````Helper functions`````````````````````````````````
def printTime(): return time.strftime("%m%d:%H%M%S")
def printi(msg): print(f'INFO_AS@{printTime()}: {msg}')
def printw(msg): print(f'WARN_AS@{printTime()}: {msg}')
def printe(msg): print(f'ERROR_AS@{printTime()}: {msg}')

def croppedText(txt, limit=200):
    if len(txt) > limit:
        txt = txt[:limit]+'...'
    return txt

def shortkey(i:int):
    """Return string with max 2 characters, mapping i (i<1296)"""
    s = string.digits + string.ascii_lowercase
    l = len(s)
    quotient,reminder = divmod(i,l)
    return s[i] if quotient==0 else s[quotient]+s[reminder]

#````````````````````````````Serializer class`````````````````````````````````
class apstrim ():
    eventExit = threading.Event()

    def __init__(self, publisher, devPars:list, sectionInterval=60.
    , compression=False, quiet=False):
        """Create the object streamer. 
        publisher is a class, providing a subscribe() method,
        devPar: list of device:parameter strings,
        sectionInterval: time between writing of the logBook sections
        compression: compression enable flag.
        """
        #printi(f'apstrim  {__version__}, sectionInterval {sectionInterval}')
        self.publisher = publisher
        self.devPars = devPars
        self.sectionInterval = sectionInterval
        self.quiet = quiet
        signal.signal(signal.SIGINT, _safeExit)
        signal.signal(signal.SIGTERM, _safeExit)
        self.headerSection = {'apstrim ':__version__
        , 'sectionInterval':sectionInterval}
        if compression:
            import lz4framed
            self.compress = lz4framed.compress
            self.headerSection['compression'] = 'lz4framed'
        else:
            self.compress = None
            self.headerSection['compression'] = 'None'
        printi(f'Header section: {self.headerSection}')
        self.lock = threading.Lock()


    def start(self, fileName:str):
        """Start streaming of data objects to logbook file
        """
        # if file exists then rename the existing file
        try:
            modificationTime = pathlib.Path(fileName).stat().st_mtime
            dt = datetime.datetime.fromtimestamp(modificationTime)
            suffix = dt.strftime('_%Y%m%d_%H%M') 
            try:    fname,ext = fileName.rsplit('.',1)
            except:    fname,ext = fileName,''
            otherName = fname + suffix + '.' + ext
            os.rename(fileName, otherName)
            printw(f'Existing file {fileName} have been renamed to {otherName}')
        except Exception as e:
            pass

        self.logbook = open(fileName, 'wb')
        self.logbook.write(msgpack.packb(self.headerSection))
        printi(f'Logbook file: {fileName} created')

        self._create_logSection()

        #printi('starting periodic thread')
        myThread = threading.Thread(target=self._serialize_section)
        myThread.start()

        self.pars = {}
        for i,pname in enumerate(self.devPars):
            devPar = tuple(pname.rsplit(':',1))
            if True:#try:
                self.publisher.subscribe(self._delivered, devPar)
            else:#except Exception as e:
                printe(f'Subscription failed for {pname}: {e}')
                continue
            self.pars[pname] = [shortkey(i)]
        printi(f'parameters: {self.pars}')
        self.logbook.write(msgpack.packb({'parameters':self.pars}))

    def stop(self):
        """Stop the streaming"""
        self.eventExit.set()
        #self.logbook.close()

    def _delivered(self, *args):
        """Callback, specified in the subscribe() request. 
        Called when the requested data have been changed.
        args is a map of delivered objects."""
        #print(f'delivered: {args}')
        timestampedMap = {}
        for devPar,props in args[0].items():
            #print(f'devPar: {devPar,props}, {type(devPar)}')
            try:
              if isinstance(devPar, tuple):
                # EPICS and ADO packing
                dev,par = devPar
                value = props['value']
                timestamp = props.get('timestamp')# valid in EPICS and LITE
                if timestamp == None:# decode ADO timestamp 
                    timestamp = int(props['timestampSeconds']*Nano
                    + props['timestampNanoSeconds'])
                else:
                    timestamp = int(timestamp*Nano)
                skey = self.pars[dev+':'+par][0]
              elif devPar == 'ppmuser':# ADO has extra item, skip it.
                continue
              else:
                #LITE packing:
                pars = props
                for par in pars:
                    try:
                        value = pars[par]['v']
                        timestamp = int(pars[par]['t'])
                    except: # try old LITE packing
                        value = pars[par]['value']                     
                        timestamp = int(pars[par]['timestamp']*Nano)
                    skey = self.pars[devPar+':'+par][0]
            except Exception as e:
                printw(f'exception in unpacking: {e}')
                continue
            if timestamp in timestampedMap:
                timestampedMap[timestamp][skey] = value
            else:
                timestampedMap[timestamp] = {skey:value}
        #TODO: timestampedMap may need sorting
        #print(f'timestampedMap: {timestampedMap}')
        with self.lock:
            self.logParagraph.append(list(timestampedMap.items())[0])
        
    def _create_logSection(self):
      with self.lock:
        self.logParagraph = []
        key = time.strftime("%y%m%d:%H%M%S")
        self.logSection = (key, self.logParagraph)

    def _serialize_section(self):
        #printi('serialize_section started')
        periodic_update = time.time()
        stat = [0, 0]
        try:
          while not self.eventExit.is_set():
            self.eventExit.wait(self.sectionInterval)
            if len(self.logSection[SecParagraph]) == 0:
                continue

            #print(f'write section {self.logSection}')
            packed = msgpack.packb(self.logSection)
            if self.compress is not None:
                compressed = self.compress(packed)
                packed = msgpack.packb(compressed)
            self.logbook.write(packed)
            self.logbook.flush()

            stat[0] += len(self.logSection[SecParagraph])
            stat[1] += len(packed)
            self._create_logSection()
            timestamp = time.time()
            dt = timestamp - periodic_update
            if dt > 10.:
                periodic_update = timestamp
                if not self.quiet:
                    print(f'{time.strftime("%y-%m-%d %H:%M:%S")} Logged {stat[0]} paragraphs, {stat[1]/1000.} KBytes')
        except Exception as e:
            print(f'ERROR: Exception in serialize_section: {e}')
        print(f'Logging finished after {stat[0]} paragraphs')
        self.logbook.close()
                
def _safeExit(_signo, _stack_frame):#, self=None):
    print('safeExit')
    apstrim.eventExit.set()
    
                



"""Test of an apstrim-generated files.
It deserealizes the scalar and vector objects from files and plots them as 
a stripchart.
"""
import sys, time, argparse
from timeit import default_timer as timer
from functools import partial
import numpy as np
import msgpack
import msgpack_numpy
msgpack_numpy.patch()
import pyqtgraph as pg

__version__ = 'v06 2021-06-14'# improved plotting: cursors, legends. Qt4 compatible.
Nano = 1000000000
#````````````````````````````Helper methods```````````````````````````````````
def decompress(arg):
    # Stub function for decompressing
    return
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,
#````````````````````````````Deserialize files````````````````````````````````
parser = argparse.ArgumentParser(description=__doc__)
#parser.add_argument('-k', '--keys', help=('Items to plot. '
#'String of 1-letter keys of the parameter map e.g. 1357'))
parser.add_argument('-f', '--fastPlotting', action='store_true', help=\
'Fast plotting (for large data sets)')
parser.add_argument('files', nargs='*', default=['apstrim.aps'], help=\
'Input files, Unix style pathname pattern expansion allowed e.g: pi0_2021*.aps')
pargs = parser.parse_args()
#print(f'pargs: {pargs}')
print(f'files: {pargs.files}')

plotData ={}
for file in pargs.files:
    print(f'Processing {file}')
    f = open(file,'rb')
    book = msgpack.Unpacker(f)
    
    nSections = 0
    nParagraphs = 0
    for section in book:
        #if nSections > 100:  break
        nSections += 1
        if nSections == 1:# skip info section 
            print(f'file info: {section}')
            compression = section.get('compression')
            if compression is None:
                continue
            if compression != 'None':
                module = __import__(compression)
                decompress = module.decompress
            continue
        if nSections == 2:# section: parameters
            par2key = section['parameters']
            key2par = {value[0]:key for key,value in par2key.items()}
            print(f'parameter map: {key2par}')
            for key,par in key2par.items():
                #if key not in plotData:
                #    plotData[key] = {'par':par, 'x':[], 'y':[]}
                if par not in plotData:
                    #print(f'add to graph[{len(plotData)+1}]: {par}') 
                    plotData[par] = {'x':[], 'y':[]}
            continue
    
        # data sections
        #print(f'dsec: {section}')
        try:
            if compression != 'None':
                decompressed = decompress(section)
                section = msgpack.unpackb(decompressed)
            sectionDatetime, paragraphs = section
            #print(f'Section time: {sectionDatetime}')
        except Exception as e:
            print(f'WARNING: wrong section {nSections}: {str(section)[:75]}...')
            continue
        nParagraphs += len(paragraphs)
        try:
          for timestamp,parkeys in paragraphs:
            timestamp /= Nano
            #print(f'paragraph: {timestamp,parkeys}')
            for key,values in parkeys.items():
                try:    nVals = len(values)
                except: values = [values] # make it subscriptable
                par = key2par[key]
                # if values is a vector then append all its points spaced by 1 us
                for i,v in enumerate(values):
                    plotData[par]['x'].append(timestamp + i*1.e-6)
                    plotData[par]['y'].append(v)
                    #print(f'key {key}, ts: {timestamp}')
        except Exception as e:
            print(f'WARNING: wrong paragraph {nParagraphs}')

    print(f'Deserialized from {file}: {nSections} sections, {nParagraphs} paragraphs')
#,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,    
#```````````````````````````` Plot objects````````````````````````````````````
class DateAxis(pg.AxisItem):
    """Time scale for plotItem"""
    def tickStrings(self, values, scale, spacing):
        strns = []
        if len(values) == 0: 
            return ''
        rng = max(values)-min(values)
        #if rng < 120:
        #    return pg.AxisItem.tickStrings(self, values, scale, spacing)
        if rng < 3600*24:
            string = '%d %H:%M:%S'
        elif rng >= 3600*24 and rng < 3600*24*30:
            string = '%d'
        elif rng >= 3600*24*30 and rng < 3600*24*30*24:
            string = '%b'
        elif rng >=3600*24*30*24:
            string = '%Y'
        for x in values:
            try:
                strns.append(time.strftime(string, time.localtime(x)))
            except ValueError:  ## Windows can't handle dates before 1970
                strns.append('')
        return strns

#QT5#win = pg.GraphicsLayoutWidget(show=True)
qApp = pg.mkQApp()
win = pg.GraphicsLayoutWidget()
win.show()

win.resize(800,600)
s = pargs.files[0] if len(pargs.files)==1 else pargs.files[0]+'...'
win.setWindowTitle(f'Graphs[{len(plotData)}] of {s}')

plotItem = win.addPlot(#title="apstrim plotItem",
    axisItems={'bottom':DateAxis(orientation='bottom')})
legend = pg.LegendItem((80,60), offset=(70,20))
legend.setParentItem(plotItem)
viewBox = plotItem.getViewBox()
viewBox.setMouseMode(pg.ViewBox.RectMode)

idx = 0
ts = timer()
nPoints = 0
viewRect = plotItem.viewRect()
xSize = int(viewRect.right() - viewRect.left())
for par,xy in plotData.items():
    print(f'Graph[{idx}]: {par}')
    idx += 1
    # sort points along X
    a = np.stack([xy['x'],xy['y']])
    a = a[:, a[0, :].argsort()]
    pen = (idx,len(plotData))
    npt = len(a[0])
    nPoints += npt
    try:
        if pargs.fastPlotting or npt/xSize > 100:
            # plotting of only lines
            p = plotItem.plot(a[0], a[1], pen=pen)
        else:
            # plotting with symbols is 10 times slower
            p = plotItem.plot(a[0], a[1], pen=pen
            ,symbol='+', symbolSize=2, symbolPen=pen)
        legend.addItem(p, par)
    except Exception as e:
        print(f'WARNING: plotting is not supported for item {par}: {e}')
print(f'plotting time of {nPoints} points: {round(timer()-ts,3)} s')

cursors = set()
def add_cursor(direction):
    global cursor
    angle = {'Vertical':90, 'Horizontal':0}[direction]
    vid = {'Vertical':0, 'Horizontal':1}[direction]
    viewRange = plotItem.viewRange()
    pos = (viewRange[vid][1] + viewRange[vid][0])/2.
    pen = pg.mkPen(color='y', width=1, style=pg.QtCore.Qt.DotLine)
    cursor = pg.InfiniteLine(pos=pos, pen=pen, movable=True, angle=angle
    , label=str(round(pos,3)))
    cursor.sigPositionChangeFinished.connect(\
    (partial(cursorPositionChanged,cursor)))
    cursors.add(cursor)
    plotItem.addItem(cursor)
    cursorPositionChanged(cursor)

def cursorPositionChanged(cursor):
    pos = cursor.value()
    horizontal = cursor.angle == 0.
    viewRange = plotItem.viewRange()[horizontal]
    if pos > viewRange[1]:
        plotItem.removeItem(cursor)
        cursors.remove(cursor)
    else:
        if horizontal:
            text = str(round(pos,3))
        else:
            text = time.strftime('%H:%M:%S', time.localtime(pos))
        cursor.label.setText(text)

add_cursor('Vertical')
add_cursor('Horizontal')

qApp.exec_()



#!/usr/bin/env python3
"""This script provide conversion of LDO name to (host:port,device)"""
# If CNSHostPort is defined, then the name will be resolved using a dedicated
#liteCNSServer. If it is not defined then the name will be resolved using 
#liteCNS.yaml file.
__version__='v01 2021-04-19'#
print(f'liteCNS version {__version__}')

# Comment the following line for file-based name resolution
#CNSHostPort = 'acnlin23;9699'# host;port of the liteServer

#`````````````````````````````````````````````````````````````````````````````
def hostPort(cnsName):
  #print('>liteCNS.hostPort(%s)'%cnsName)
  try:
    CNSHostPort
    # CNSHostPort is defined, use liteCNSServer
    raise NameError('ERROR:CNS, liteCNSServer is not supported yet')
    
  except NameError: # file-based name resolution
    # CNSHostPort is not defined, use liteCNS.yaml
    import yaml
    fname = '/operations/app_store/liteServer/liteCNS.yaml'
    f = open(fname,'r')
    print('INFO.CNS: File-base name resolution using '+fname)
    y = yaml.load(f,Loader=yaml.BaseLoader)
    #print(f'liteCNS: {y}')
    try:
        r = y['hosts'][cnsName]
    except Exception as e:
        raise NameError('ERROR:CNS '+str(e))
    return r




# Copyright (c) 2021 Andrei Sukhanov. All rights reserved.
#
# Licensed under the MIT License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://github.com/ASukhanov/apstrim/blob/main/LICENSE
#

"""Basic EPICS access API using caproto: a pure-Python Channel Access protocol library
"""
__version__ = 'v02 2021-06-01'# 
print(f'{__file__}, {__version__}')

#from timeit import default_timer as timer

from caproto.threading.client import Context
Ctx = Context()

PVC_PV, PVC_CB, PVC_Props = 0, 1, 2
dbg = True
CA_data_type_STRING = 14

def printd(msg):
    if dbg:
        print(f'CPAccess:{msg}')
    
class Access():
    __version__ = __version__
    Dbg = False

    def __init__(self):
        self.PVCache = {}# cache of PVs
        self.subscriptions = []

    def _get_pv(self, pvName):
        r = self.PVCache.get(pvName)
        if r is not None:
            pv,*_ = r
        else:
            #print( f'register pv {pvName}')
            pv,*_ = Ctx.get_pvs(pvName, timeout=2)
            #print(f'>pvCache: {pv}')
            self.PVCache[pvName] = [pv, None, None]
            self._fill_PVCacheProps(pvName)
        return pv
    
    def _fill_PVCacheProps(self, pvName):
        pv = self.PVCache[pvName][PVC_PV]
        if pv is None:
            return None
        pvData = pv.read(data_type='time')
        #printd(f'pvData:{pvData}')
        val = pvData.data
        #printd(f'val:{val}')
        if len(val) == 1:
            try:
                # treat it as numpy
                val = pvData.data[0].item()
            except:
                # data is not numpy
                val = pvData.data[0]
        
        # get properties
        #ISSUE: the caproto reports timestamp as float, the precision for float64 
        #presentation is ~300ns
        featureBit2Letter = {1:'R', 2:'WE'}
        featureCode = pv.access_rights
        features = ''
        for bit, letter in featureBit2Letter.items():
            if bit & featureCode:
                features += letter
        pvControl = pv.read(data_type='control')
        #printd(f'pvcontrol {pvName}: {pvControl}')
        datatype = pvControl.data_type
        #printd(f'data_type:{datatype}')
        if datatype == CA_data_type_STRING:# convert text bytearray to str
            val = val.decode()
        props = {'value':val}
        props['timestamp'] = pvData.metadata.timestamp
        props['count'] = len(pvData.data)
        props['features'] = features
        try:    
            props['units'] = pvControl.metadata.units.decode()
            if props['units'] == '':   props['units'] = None
        except: pass
    
        try:    props['engLow'] = pvControl.metadata.lower_ctrl_limit
        except: pass
    
        try:    
            props['engHigh'] = pvControl.metadata.upper_ctrl_limit
            if props['engHigh'] == 0.0 and props['engHigh'] == 0.0:
                props['engHigh'], props['engLow'] = None, None
        except: pass
    
        try:
            props['alarm'] = pvControl.metadata.severity 
            #printd(f'status {pvControl.metadata.severity}')
            if props['alarm'] == 17:# UDF
                props['alarm'] = None
        except: pass
    
        try:    # legalValues
            enum_strings = pvControl.metadata.enum_strings
            props['legalValues'] = [i.decode() for i in enum_strings]
            props['value'] = props['legalValues'][val]
        except:
            #props['legalValues'] = None
            if 'legalValues' in props:
                del props['legalValues']
        #printd(f'_props {props}')
        self.PVCache[pvName][PVC_Props] = props
    
    def _unpack_ReadNotifyResponse(self, pvName, pvData):
        #printd('>uRNR')
        val = pvData.data
        if len(val) == 1:
            try:    #it as numpy
                val = pvData.data[0].item()
            except: # it is not numpy
                val = pvData.data[0]
        #printd(f'pvData:{pvData}')
        #printd(f'val:{val}, {pvData.data_type}')
        if pvData.data_type == CA_data_type_STRING:
            val = val.decode()
        rDict = {'pvname':pvName, 'value':val}
        
        rDict['timestamp'] = pvData.metadata.timestamp
        ##printd(f'uRNR1:{rDict}')
    
        legalValues = self.PVCache[pvName][PVC_Props].get('legalValues')
        #printd(f'uRNR2:{legalValues}')
        if legalValues is not None:
            #rDict['value'] = legalValues[int(val)]
            val = legalValues[int(val)]
        #printd('uRNR3')
        alarm = pvData.metadata.severity
        rDict['alarm'] = alarm
        #printd(f'pvName:{pvName}')
        key = tuple(pvName.rsplit(':',1))
        #printd(f'key:{key}')
        rDict = {key: {'value':val\
        , 'timestamp':pvData.metadata.timestamp, 'alarm': alarm}}
        #printd(f'<uRNR:{rDict}')
        return rDict
    
    def info(self, devParName):
        """Abridged PV info"""
        pvName = ':'.join(devParName)
        pv = self._get_pv(pvName)
        return {pvName:self.PVCache[pvName][PVC_Props]}


    def get(self, devParName, **kwargs):
        pvName = ':'.join(devParName)
        pv = self._get_pv(pvName)
        pvData = pv.read(data_type='time')
        rDict = self._unpack_ReadNotifyResponse(pvName, pvData)
        return rDict

    def set(self, devParValue):
        dev, par, value = devParValue
        #print(f'epicsAccess.set({dev,par,value})')
        pvName = ':'.join((dev,par))
        pv = self._get_pv(pvName)
        try: # if PV has legalValues then the value should be index of legalValues
            value = self.PVCache[pvName][PVC_Props]['legalValues'].index(value)
            #lv = self.PVCache[pvName][PVC_Props]['legalValues']
            #print(f'lv:{lv}')
        except Exception as e:
            #print(f'in epicsAccess.set. Value not in legalValues: {e}')
            pass
        pv.write(value)
        return 1

    def _callback(self, subscription, pvData):
        #print(f'>epicsAccess._callback: {pvData})')
        #tMark = [timer(), 0., 0.]
        pvName = subscription.pv.name
        rDict = self._unpack_ReadNotifyResponse(pvName, pvData)
        #tMark[1] = timer()
        #printd(f'rDict for {pvName}:{rDict}')
        #printd(f'self.PVCache:{self.PVCache}')
        cache = self.PVCache.get(pvName)
        #printd(f'cache[{len(cache)}]:{cache}')
        cb = cache[PVC_CB]
        #printd(f'cb:{str(cb)}')
        if cb:
            #print(f'epicsAccess call {cb}({rDict})')
            cb(rDict)
        #tMark[2] = timer() - tMark[1]
        #tMark[1] -= tMark[0]
        ##printd(f'caproto cb times {tMark}')# 20-30 uS
        #printd('<callback')
        
    def subscribe(self, callback, devParName):
        pvName = ':'.join(devParName)
        if not isinstance(pvName, str):
            msg = f'ERROR: Second argument of subscribe() should be a string, not {type(pvName)}'
            raise SystemError(msg)
        pv = self._get_pv(pvName)
        subscription = pv.subscribe(data_type='time')
        self.PVCache[pvName][PVC_CB] = callback
        #printd('>add_callback')
        subscription.add_callback(self._callback)
        self.subscriptions.append(subscription)
    
    def unsubscribe(self):
        for subscription in self.subscriptions:
            #print(f'>epicsAccess clear subs: {subscription}')
            self.subscriptions.clear()
        self.subscriptions = []
    



Description = '''Serializer of Process Variables (from EPICS infrastructure)
or Data Objects from other infrastructures, e.g. ADO or LITE).'''

import sys, argparse
from .apstrim  import apstrim 

def main():
    
    # parse common arguments
    parser = argparse.ArgumentParser(description=Description)
    parser.add_argument('-c', '--compress', action='store_true', help=\
    'Enable online compression')
    #parser.add_argument('-d', '--dbg', action='store_true', help=\
    #'turn on debugging')
    #parser.add_argument('-f', '--file', default=None, help=\
    #'Configuration file')
    parser.add_argument('-o', '--outfile', default='apstrim.aps', help=\
    'Logbook file for storing PVs and data objects')
    parser.add_argument('-n', '--namespace', default='EPICS', help=\
    'Infrastructure namespace, e.g.: EPICS, ADO or LITE')
    parser.add_argument('-t', '--sectionTime', type=float, default=60., help=\
    'Time interval of writing of sections to logbook')
    parser.add_argument('-q', '--quiet', action='store_true', help=\
    'Quiet: dont print section progress')
    parser.add_argument('pvNames', nargs='*', help=\
    'Data Object names, one item per device parameters are comma-separated: dev1:par1,par2 dev2:par1,par2') 
    pargs = parser.parse_args()
    #print(f'pargs:{pargs}')

    s = pargs.namespace.upper()
    if s == 'LITE':
        from .pubLITE import Access as publisher
    elif s == 'EPICS':
        from .pubEPICS import Access
        publisher = Access()
    else:
        print(f'ERROR: Unsupported namespace {s}')
        sys.exit(1)

    pvNames = []
    for pvn in pargs.pvNames:
        tokens = pvn.split(',')
        first = tokens[0]
        pvNames.append(first)
        if len(tokens) > 1:
            dev = first.rsplit(':',1)[0]
            for par in tokens[1:]:
                pvNames.append(dev+':'+par)
    #print(f'pvNames: {pvNames}')

    aps = apstrim(publisher, pvNames, pargs.sectionTime, pargs.compress
    , pargs.quiet)
    aps.start(pargs.outfile)

    timeSpan = 0 #pargs.sectionTime*2 + 5
    txt = f'for {round(timeSpan/60., 2)} m' if timeSpan else 'endlessly'
    print(f'Streaming started {txt}, press Ctrl/C to stop.')
    if timeSpan:
        aps.eventExit.wait(timeSpan)
        aps.stop()

if __name__ == '__main__':
    main()
